from secator.decorators import task
from secator.definitions import (DELAY, DEPTH, FILTER_CODES, FILTER_REGEX,
							   FILTER_SIZE, FILTER_WORDS, FOLLOW_REDIRECT,
							   HEADER, MATCH_CODES, MATCH_REGEX, MATCH_SIZE,
							   MATCH_WORDS, METHOD, OPT_NOT_SUPPORTED,
							   OPT_PIPE_INPUT, PROXY, RATE_LIMIT, RETRIES,
							   THREADS, TIMEOUT, USER_AGENT)
from secator.serializers import JSONSerializer
from secator.tasks._categories import HttpCrawler


@task()
class gau(HttpCrawler):
	"""Fetch known URLs from AlienVault's Open Threat Exchange, the Wayback Machine, Common Crawl, and URLScan."""
	cmd = 'gau'
	file_flag = OPT_PIPE_INPUT
	json_flag = '--json'
	opt_prefix = '--'
	opts = {
		'providers': {'type': str, 'default': None, 'help': 'List of providers to use (wayback,commoncrawl,otx,urlscan)'}
	}
	opt_key_map = {
		HEADER: OPT_NOT_SUPPORTED,
		DELAY: OPT_NOT_SUPPORTED,
		DEPTH: OPT_NOT_SUPPORTED,
		FILTER_CODES: 'fc',
		FILTER_REGEX: OPT_NOT_SUPPORTED,
		FILTER_SIZE: OPT_NOT_SUPPORTED,
		FILTER_WORDS: OPT_NOT_SUPPORTED,
		MATCH_CODES: 'mc',
		MATCH_REGEX: OPT_NOT_SUPPORTED,
		MATCH_SIZE: OPT_NOT_SUPPORTED,
		MATCH_WORDS: OPT_NOT_SUPPORTED,
		FOLLOW_REDIRECT: OPT_NOT_SUPPORTED,
		METHOD: OPT_NOT_SUPPORTED,
		PROXY: 'proxy',
		RATE_LIMIT: OPT_NOT_SUPPORTED,
		RETRIES: 'retries',
		THREADS: 'threads',
		TIMEOUT: 'timeout',
		USER_AGENT: OPT_NOT_SUPPORTED,
	}
	item_loaders = [JSONSerializer()]
	install_pre = {
		'apk': ['libc6-compat']
	}
	install_cmd = 'go install -v github.com/lc/gau/v2/cmd/gau@latest'
	install_github_handle = 'lc/gau'
	proxychains = False
	proxy_socks5 = True
	proxy_http = True
	profile = 'io'



import operator

import yaml
from rich.console import Console
from rich.table import Table

console = Console(stderr=True)
console_stdout = Console(record=True)
# handler = RichHandler(rich_tracebacks=True)  # TODO: add logging handler


def criticity_to_color(value):
	if value == 'critical':
		value = f'[bold red3]{value.upper()}[/]'
	elif value == 'high':
		value = f'[bold orange_red1]{value.upper()}[/]'
	elif value == 'medium':
		value = f'[bold dark_orange]{value.upper()}[/]'
	elif value == 'low':
		value = f'[bold yellow1]{value.upper()}[/]'
	elif value == 'info':
		value = f'[bold green]{value.upper()}[/]'
	return value


def status_to_color(value):
	value = int(value) if value else None
	if value is None:
		return value
	if value < 400:
		value = f'[bold green]{value}[/]'
	elif value in [400, 499]:
		value = f'[bold dark_orange]{value}[/]'
	elif value >= 500:
		value = f'[bold red3]{value}[/]'
	return value


FORMATTERS = {
	'confidence': criticity_to_color,
	'severity': criticity_to_color,
	'cvss_score': lambda score: '' if score == -1 else f'[bold cyan]{score}[/]',
	'port': lambda port: f'[bold cyan]{port}[/]',
	'url': lambda host: f'[bold underline blue]{host}[/]',
	'ip': lambda ip: f'[bold yellow]{ip}[/]',
	'status_code': status_to_color,
	'reference': lambda reference: f'[link={reference}]🡕[/]',
	'_source': lambda source: f'[bold gold3]{source}[/]'
}


def build_table(items, output_fields=[], exclude_fields=[], sort_by=None):
	"""Build rich table.

	Args:
		items (list): List of items.
		output_fields (list, Optional): List of fields to add.
		exclude_fields (list, Optional): List of fields to exclude.
		sort_by (tuple, Optional): Tuple of sort_by keys.

	Returns:
		rich.table.Table: rich table.
	"""
	# Sort items by one or multiple fields
	if sort_by and all(sort_by):
		items = sorted(items, key=operator.attrgetter(*sort_by))

	# Create rich table
	table = Table(show_lines=True)

	# Get table schema if any, default to first item keys
	keys = []
	if output_fields:
		keys = [k for k in output_fields if k not in exclude_fields]
		# Remove meta fields not needed in output
		if '_cls' in keys:
			keys.remove('_cls')
		if '_type' in keys:
			keys.remove('_type')
		if '_uuid' in keys:
			keys.remove('_uuid')

		# Add _source field
		if '_source' not in keys:
			keys.append('_source')

		# Create table columns
		for key in keys:
			key_str = key
			if not key.startswith('_'):
				key_str = ' '.join(key.split('_')).title()
			no_wrap = key in ['url', 'reference', 'references', 'matched_at']
			overflow = None if no_wrap else 'fold'
			table.add_column(
				key_str,
				overflow=overflow,
				min_width=10,
				no_wrap=no_wrap)

	if not keys:
		table.add_column(
			'Extracted values',
			overflow=False,
			min_width=10,
			no_wrap=False)

	# Create table rows
	for item in items:
		values = []
		if keys:
			for key in keys:
				value = getattr(item, key) if keys else item
				value = FORMATTERS.get(key, lambda x: x)(value) if keys else item
				if isinstance(value, dict) or isinstance(value, list):
					value = yaml.dump(value)
				elif isinstance(value, int) or isinstance(value, float):
					value = str(value)
				values.append(value)
		else:
			values = [item]
		table.add_row(*values)
	return table



import time
from dataclasses import dataclass, field
from enum import Enum

from secator.definitions import ALIVE, IP
from secator.output_types import OutputType
from secator.utils import rich_to_ansi


class IpProtocol(str, Enum):
	IPv6 = 'IPv6'
	IPv4 = 'IPv4'


@dataclass
class Ip(OutputType):
	ip: str
	host: str = ''
	alive: bool = False
	protocol: str = field(default=IpProtocol.IPv4)
	_source: str = field(default='', repr=True)
	_type: str = field(default='ip', repr=True)
	_timestamp: int = field(default_factory=lambda: time.time(), compare=False)
	_uuid: str = field(default='', repr=True, compare=False)
	_context: dict = field(default_factory=dict, repr=True, compare=False)
	_tagged: bool = field(default=False, repr=True, compare=False)
	_duplicate: bool = field(default=False, repr=True, compare=False)
	_related: list = field(default_factory=list, compare=False)

	_table_fields = [IP, ALIVE]
	_sort_by = (IP,)

	def __str__(self) -> str:
		return self.ip

	def __repr__(self) -> str:
		s = f'💻 [bold white]{self.ip}[/]'
		if self.host:
			s += rf' \[[bold magenta]{self.host}[/]]'
		return rich_to_ansi(s)



from click.testing import CliRunner

from secator.cli import cli

import unittest


class TestCli(unittest.TestCase):

	def setUp(self):
		self.runner = CliRunner()

	def test_cli_version(self):
		result = self.runner.invoke(cli, ['--version'])
		assert not result.exception
		assert result.exit_code == 0
		assert 'Current version' in result.output

	def test_task_command(self):
		result = self.runner.invoke(cli, ['task'])
		assert not result.exception
		assert result.exit_code == 0
		assert 'Run a task.' in result.output

	def test_workflow_command(self):
		result = self.runner.invoke(cli, ['workflow'])
		assert not result.exception
		assert result.exit_code == 0
		assert 'Run a workflow.' in result.output

	def test_scan_command(self):
		result = self.runner.invoke(cli, ['scan'])
		assert not result.exception
		assert result.exit_code == 0
		assert 'Run a scan.' in result.output

	# def test_worker_command(self):
	# 	result = self.runner.invoke(cli, ['worker', '--check'])
	# 	assert not result.exception
	# 	assert result.exit_code == 0
	# 	# Add more assertions based on expected output

	def test_util_proxy_command(self):
		result = self.runner.invoke(cli, ['util', 'proxy', '--timeout', '0.5', '--number', '2'])
		assert not result.exception
		assert result.exit_code == 0
		# Add more assertions based on expected output

	def test_util_revshell_command(self):
		result = self.runner.invoke(cli, ['util', 'revshell', '--host', '127.0.0.1', '--port', '9001'])
		assert not result.exception
		assert result.exit_code == 0
		# Add more assertions based on expected output

	# def test_util_serve_command(self):
	# 	result = self.runner.invoke(cli, ['util', 'serve', '--directory', '.', '--port', '8000'])
	# 	assert not result.exception
	# 	assert result.exit_code == 0
	# 	# Add more assertions based on expected output

	def test_config_get_command(self):
		result = self.runner.invoke(cli, ['config', 'get'])
		assert not result.exception
		assert result.exit_code == 0
		# Add more assertions based on expected output

	def test_config_set_command(self):
		result = self.runner.invoke(cli, ['config', 'set', 'key', 'value'])
		assert not result.exception
		assert result.exit_code == 0
		# Add more assertions based on expected output

	def test_report_show_command(self):
		result = self.runner.invoke(cli, ['report', 'show'])
		assert not result.exception
		assert result.exit_code == 0
		# Add more assertions based on expected output

	def test_install_addons_worker_command(self):
		result = self.runner.invoke(cli, ['install', 'addons', 'worker'])
		assert not result.exception
		assert result.exit_code == 0
		# Add more assertions based on expected output

	def test_health_command(self):
		result = self.runner.invoke(cli, ['health'])
		assert not result.exception
		assert result.exit_code == 0
		# Add more assertions based on expected output

	# Add more tests for other commands as needed

if __name__ == '__main__':
	unittest.main()



#!/usr/bin/python

import os

from importlib.metadata import version

from secator.config import CONFIG, ROOT_FOLDER


# Globals
VERSION = version('secator')
ASCII = rf"""
			 __            
   ________  _________ _/ /_____  _____
  / ___/ _ \/ ___/ __ `/ __/ __ \/ ___/
 (__  /  __/ /__/ /_/ / /_/ /_/ / /    
/____/\___/\___/\__,_/\__/\____/_/     v{VERSION}

			freelabz.com
"""  # noqa: W605,W291

# Debug
DEBUG = CONFIG.debug.level
DEBUG_COMPONENT = CONFIG.debug.component.split(',')

# Default tasks settings
DEFAULT_NUCLEI_FLAGS = os.environ.get('DEFAULT_NUCLEI_FLAGS', '-stats -sj -si 20 -hm -or')
DEFAULT_FEROXBUSTER_FLAGS = os.environ.get('DEFAULT_FEROXBUSTER_FLAGS', '--auto-bail --no-state')

# Constants
OPT_NOT_SUPPORTED = -1
OPT_PIPE_INPUT = -1
STATE_COLORS = {
	'PENDING': 'dim yellow3',
	'RUNNING': 'bold yellow3',
	'SUCCESS': 'bold green',
	'FAILURE': 'bold red',
	'REVOKED': 'bold magenta'
}

# Vocab
ALIVE = 'alive'
AUTO_CALIBRATION = 'auto_calibration'
CONTENT_TYPE = 'content_type'
CONTENT_LENGTH = 'content_length'
CIDR_RANGE = 'cidr_range'
CPES = 'cpes'
CVES = 'cves'
DELAY = 'delay'
DOMAIN = 'domain'
DEPTH = 'depth'
EXTRA_DATA = 'extra_data'
EMAIL = 'email'
FILTER_CODES = 'filter_codes'
FILTER_WORDS = 'filter_words'
FOLLOW_REDIRECT = 'follow_redirect'
FILTER_REGEX = 'filter_regex'
FILTER_SIZE = 'filter_size'
HEADER = 'header'
HOST = 'host'
IP = 'ip'
PROTOCOL = 'protocol'
LINES = 'lines'
METHOD = 'method'
MATCH_CODES = 'match_codes'
MATCH_REGEX = 'match_regex'
MATCH_SIZE = 'match_size'
MATCH_WORDS = 'match_words'
OUTPUT_PATH = 'output_path'
PATH = 'path'
PERCENT = 'percent'
PORTS = 'ports'
PORT = 'port'
PROXY = 'proxy'
RATE_LIMIT = 'rate_limit'
RETRIES = 'retries'
TAGS = 'tags'
THREADS = 'threads'
TIME = 'time'
TIMEOUT = 'timeout'
TOP_PORTS = 'top_ports'
TYPE = 'type'
URL = 'url'
USER_AGENT = 'user_agent'
USERNAME = 'username'
STORED_RESPONSE_PATH = 'stored_response_path'
SCRIPT = 'script'
SERVICE_NAME = 'service_name'
SOURCES = 'sources'
STATE = 'state'
STATUS_CODE = 'status_code'
TECH = 'tech'
TITLE = 'title'
SITE_NAME = 'site_name'
SERVICE_NAME = 'service_name'
CONFIDENCE = 'confidence'
CVSS_SCORE = 'cvss_score'
DESCRIPTION = 'description'
ID = 'id'
MATCHED_AT = 'matched_at'
NAME = 'name'
PROVIDER = 'provider'
REFERENCE = 'reference'
REFERENCES = 'references'
SEVERITY = 'severity'
TAGS = 'tags'
WEBSERVER = 'webserver'
WORDLIST = 'wordlist'
WORDS = 'words'


def is_importable(module_to_import):
	import importlib
	try:
		importlib.import_module(module_to_import)
		return True
	except ModuleNotFoundError:
		return False
	except Exception as e:
		print(f'Failed trying to import {module_to_import}: {str(e)}')
		return False


ADDONS_ENABLED = {}

for addon, module in [
	('worker', 'eventlet'),
	('gdrive', 'gspread'),
	('gcs', 'google.cloud.storage'),
	('mongodb', 'pymongo'),
	('redis', 'redis'),
	('dev', 'flake8'),
	('trace', 'memray'),
	('build', 'hatch')
]:
	ADDONS_ENABLED[addon] = is_importable(module)

# Check dev package
if os.path.exists(f'{ROOT_FOLDER}/pyproject.toml'):
	DEV_PACKAGE = True
else:
	DEV_PACKAGE = False



import time
from dataclasses import dataclass, field
from typing import List

from secator.definitions import DOMAIN, HOST, SOURCES
from secator.output_types import OutputType
from secator.utils import rich_to_ansi


@dataclass
class Subdomain(OutputType):
	host: str
	domain: str
	sources: List[str] = field(default_factory=list, compare=False)
	extra_data: dict = field(default_factory=dict, compare=False)
	_source: str = field(default='', repr=True)
	_type: str = field(default='subdomain', repr=True)
	_timestamp: int = field(default_factory=lambda: time.time(), compare=False)
	_uuid: str = field(default='', repr=True, compare=False)
	_context: dict = field(default_factory=dict, repr=True, compare=False)
	_tagged: bool = field(default=False, repr=True, compare=False)
	_duplicate: bool = field(default=False, repr=True, compare=False)
	_related: list = field(default_factory=list, compare=False)

	_table_fields = [
		HOST,
		DOMAIN,
		SOURCES
	]
	_sort_by = (HOST,)

	def __str__(self):
		return self.host

	def __repr__(self):
		sources_str = ', '.join([f'[magenta]{source}[/]' for source in self.sources])
		s = f'🏰 [white]{self.host}[/]'
		if sources_str:
			s += f' [{sources_str}]'
		if self.extra_data:
			s += r' \[[bold yellow]' + ', '.join(f'{k}:{v}' for k, v in self.extra_data.items()) + '[/]]'
		return rich_to_ansi(s)



import re

from secator.config import CONFIG
from secator.decorators import task
from secator.definitions import (CVES, EXTRA_DATA, ID, MATCHED_AT, NAME,
								 PROVIDER, REFERENCE, TAGS, OPT_NOT_SUPPORTED)
from secator.output_types import Exploit
from secator.runners import Command
from secator.serializers import JSONSerializer


SEARCHSPLOIT_TITLE_REGEX = re.compile(r'^((?:[a-zA-Z\-_!\.()]+\d?\s?)+)\.?\s*(.*)$')


@task()
class searchsploit(Command):
	"""Exploit searcher based on ExploitDB."""
	cmd = 'searchsploit'
	input_flag = None
	json_flag = '--json'
	version_flag = OPT_NOT_SUPPORTED
	opts = {
		'strict': {'short': 's', 'is_flag': True, 'default': False, 'help': 'Strict match'}
	}
	opt_key_map = {}
	item_loaders = [JSONSerializer()]
	output_types = [Exploit]
	output_map = {
		Exploit: {
			NAME: 'Title',
			ID: 'EDB-ID',
			PROVIDER: lambda x: 'EDB',
			CVES: lambda x: [c for c in x['Codes'].split(';') if c.startswith('CVE-')],
			REFERENCE: lambda x: f'https://exploit-db.com/exploits/{x["EDB-ID"]}',
			TAGS: lambda x: searchsploit.tags_extractor(x),
			EXTRA_DATA: lambda x: {
				k.lower().replace('date_', ''): v for k, v in x.items() if k not in ['Title', 'EDB-ID', 'Codes', 'Tags', 'Source'] and v != ''  # noqa: E501
			}
		}
	}
	install_pre = {
		'apk': ['ncurses']
	}
	install_cmd = (
		f'git clone  --depth 1 --single-branch https://gitlab.com/exploit-database/exploitdb.git {CONFIG.dirs.share}/exploitdb || true && '  # noqa: E501
		f'ln -sf $HOME/.local/share/exploitdb/searchsploit {CONFIG.dirs.bin}/searchsploit'
	)
	proxychains = False
	proxy_socks5 = False
	proxy_http = False
	input_chunk_size = 1
	profile = 'io'

	@staticmethod
	def tags_extractor(item):
		tags = []
		for tag in item['Tags'].split(','):
			_tag = '_'.join(
				tag.lower().replace('-', '_',).replace('(', '').replace(')', '').split(' ')
			)
			if not _tag:
				continue
			tags.append(tag)
		return tags

	@staticmethod
	def before_init(self):
		if len(self.inputs) == 0:
			return
		_in = self.inputs[0]
		self.matched_at = None
		if '~' in _in:
			split = _in.split('~')
			self.matched_at = split[0]
			self.inputs[0] = split[1]
		self.inputs[0] = self.inputs[0].replace('httpd', '').replace('/', ' ')

	@staticmethod
	def on_item_pre_convert(self, item):
		if self.matched_at:
			item[MATCHED_AT] = self.matched_at
		return item

	@staticmethod
	def on_item(self, item):
		match = SEARCHSPLOIT_TITLE_REGEX.match(item.name)
		# if not match:
		# 	self._print(f'[bold red]{item.name} ({item.reference}) did not match SEARCHSPLOIT_TITLE_REGEX. Please report this issue.[/]')  # noqa: E501
		if match:
			group = match.groups()
			product = '-'.join(group[0].strip().split(' '))
			if len(group[1]) > 1:
				try:
					versions, title = tuple(group[1].split(' - '))
					item.name = title
					product_info = [f'{product.lower()} {v.strip()}' for v in versions.split('/')]
					item.tags = product_info + item.tags
				except ValueError:
					item.name = item.name.split(' - ')[-1]
					item.tags = [product.lower()]
					pass
			# else:
			# 	self._print(f'[bold red]{item.name} ({item.reference}) did not quite match SEARCHSPLOIT_TITLE_REGEX. Please report this issue.[/]')  # noqa: E501
		input_tag = '-'.join(self.inputs[0].replace('\'', '').split(' '))
		item.tags = [input_tag] + item.tags
		return item



import gc
import json
import logging
import os
import sys
import uuid

from time import time

from celery import Celery, chord
from celery.app import trace

from rich.logging import RichHandler
from retry import retry

from secator.celery_signals import setup_handlers
from secator.config import CONFIG
from secator.output_types import Info
from secator.rich import console
from secator.runners import Scan, Task, Workflow
from secator.utils import (debug, deduplicate, flatten, should_update)

IN_CELERY_WORKER_PROCESS = sys.argv and ('secator.celery.app' in sys.argv or 'worker' in sys.argv)

#---------#
# Logging #
#---------#

rich_handler = RichHandler(rich_tracebacks=True)
rich_handler.setLevel(logging.INFO)
logging.basicConfig(
	level='NOTSET',
	format="%(threadName)s:%(message)s",
	datefmt="[%X]",
	handlers=[rich_handler],
	force=True)
logging.getLogger('kombu').setLevel(logging.ERROR)
logging.getLogger('celery').setLevel(logging.INFO if CONFIG.debug.level > 6 else logging.WARNING)
logger = logging.getLogger(__name__)
trace.LOG_SUCCESS = "Task %(name)s[%(id)s] succeeded in %(runtime)ss"


#------------#
# Celery app #
#------------#

app = Celery(__name__)
app.conf.update({
	# Content types
	'accept_content': ['application/x-python-serialize', 'application/json'],

	# Broker config
	'broker_url': CONFIG.celery.broker_url,
	'broker_transport_options': json.loads(CONFIG.celery.broker_transport_options) if CONFIG.celery.broker_transport_options else {  # noqa: E501
		'data_folder_in': CONFIG.dirs.celery_data,
		'data_folder_out': CONFIG.dirs.celery_data,
		'control_folder': CONFIG.dirs.celery_data,
		'visibility_timeout': CONFIG.celery.broker_visibility_timeout,
	},
	'broker_connection_retry_on_startup': True,
	'broker_pool_limit': CONFIG.celery.broker_pool_limit,
	'broker_connection_timeout': CONFIG.celery.broker_connection_timeout,

	# Result backend config
	'result_backend': CONFIG.celery.result_backend,
	'result_expires': CONFIG.celery.result_expires,
	'result_backend_transport_options': json.loads(CONFIG.celery.result_backend_transport_options) if CONFIG.celery.result_backend_transport_options else {},  # noqa: E501
	'result_extended': True,
	'result_backend_thread_safe': True,
	'result_serializer': 'pickle',

	# Task config
	'task_acks_late': CONFIG.celery.task_acks_late,
	'task_compression': 'gzip',
	'task_create_missing_queues': True,
	'task_eager_propagates': False,
	'task_reject_on_worker_lost': CONFIG.celery.task_reject_on_worker_lost,
	'task_routes': {
		'secator.celery.run_workflow': {'queue': 'celery'},
		'secator.celery.run_scan': {'queue': 'celery'},
		'secator.celery.run_task': {'queue': 'celery'},
		'secator.celery.forward_results': {'queue': 'results'},
		'secator.hooks.mongodb.*': {'queue': 'mongodb'}
	},
	'task_store_eager_result': True,
	'task_send_sent_event': CONFIG.celery.task_send_sent_event,
	'task_serializer': 'pickle',

	# Worker config
	# 'worker_direct': True,  # TODO: consider enabling this to allow routing to specific workers
	'worker_max_tasks_per_child': CONFIG.celery.worker_max_tasks_per_child,
	# 'worker_max_memory_per_child': 100000  # TODO: consider enabling this
	'worker_pool_restarts': True,
	'worker_prefetch_multiplier': CONFIG.celery.worker_prefetch_multiplier,
	'worker_send_task_events': CONFIG.celery.worker_send_task_events
})
app.autodiscover_tasks(['secator.hooks.mongodb'], related_name=None)
if IN_CELERY_WORKER_PROCESS:
	setup_handlers()


@retry(Exception, tries=3, delay=2)
def update_state(celery_task, task, force=False):
	"""Update task state to add metadata information."""
	if not IN_CELERY_WORKER_PROCESS:
		return
	if not force and not should_update(CONFIG.runners.backend_update_frequency, task.last_updated_celery):
		return
	task.last_updated_celery = time()
	debug(
		'',
		sub='celery.state',
		id=celery_task.request.id,
		obj={task.unique_name: task.status, 'count': task.self_findings_count},
		obj_after=False,
		verbose=True
	)
	return celery_task.update_state(
		state='RUNNING',
		meta=task.celery_state
	)


def revoke_task(task_id, task_name=None):
	message = f'Revoked task {task_id}'
	if task_name:
		message += f' ({task_name})'
	app.control.revoke(task_id, terminate=True)
	console.print(Info(message=message))


#--------------#
# Celery tasks #
#--------------#


def chunker(seq, size):
	return (seq[pos:pos + size] for pos in range(0, len(seq), size))


@app.task(bind=True)
def handle_runner_error(self, results, runner):
	"""Handle errors in Celery workflows (chunked tasks or runners)."""
	results = forward_results(results)
	runner.results = results
	runner.log_results()
	runner.run_hooks('on_end')
	return runner.results


def break_task(task, task_opts, targets, results=[], chunk_size=1):
	"""Break a task into multiple of the same type."""
	chunks = targets
	if chunk_size > 1:
		chunks = list(chunker(targets, chunk_size))
	debug(
		'',
		obj={task.unique_name: 'CHUNKED', 'chunk_size': chunk_size, 'chunks': len(chunks), 'target_count': len(targets)},
		obj_after=False,
		sub='celery.state',
		verbose=True
	)

	# Clone opts
	opts = task_opts.copy()

	# Build signatures
	sigs = []
	task.ids_map = {}
	for ix, chunk in enumerate(chunks):
		if not isinstance(chunk, list):
			chunk = [chunk]
		if len(chunks) > 0:  # add chunk to task opts for tracking chunks exec
			opts['chunk'] = ix + 1
			opts['chunk_count'] = len(chunks)
		task_id = str(uuid.uuid4())
		opts['has_parent'] = True
		opts['enable_duplicate_check'] = False
		opts['results'] = results
		sig = type(task).si(chunk, **opts).set(queue=type(task).profile, task_id=task_id)
		full_name = f'{task.name}_{ix + 1}'
		task.add_subtask(task_id, task.name, f'{task.name}_{ix + 1}')
		info = Info(message=f'Celery chunked task created: {task_id}', _source=full_name, _uuid=str(uuid.uuid4()))
		task.add_result(info)
		sigs.append(sig)

	# Build Celery workflow
	workflow = chord(
		tuple(sigs),
		handle_runner_error.s(runner=task).set(queue='results')
	)
	return workflow


@app.task(bind=True)
def run_task(self, args=[], kwargs={}):
	print('run task')
	console.print(Info(message=f'Running task {self.request.id}'))
	kwargs['context']['celery_id'] = self.request.id
	task = Task(*args, **kwargs)
	task.run()


@app.task(bind=True)
def run_workflow(self, args=[], kwargs={}):
	console.print(Info(message=f'Running workflow {self.request.id}'))
	kwargs['context']['celery_id'] = self.request.id
	workflow = Workflow(*args, **kwargs)
	workflow.run()


@app.task(bind=True)
def run_scan(self, args=[], kwargs={}):
	console.print(Info(message=f'Running scan {self.request.id}'))
	if 'context' not in kwargs:
		kwargs['context'] = {}
	kwargs['context']['celery_id'] = self.request.id
	scan = Scan(*args, **kwargs)
	scan.run()


@app.task(bind=True)
def run_command(self, results, name, targets, opts={}):
	if IN_CELERY_WORKER_PROCESS:
		opts.update({'print_item': True, 'print_line': True, 'print_cmd': True})
		routing_key = self.request.delivery_info['routing_key']
		console.print(Info(message=f'Task "{name}" running with routing key "{routing_key}"'))

	# Flatten + dedupe + filter results
	results = forward_results(results)

	# Set Celery request id in context
	context = opts.get('context', {})
	context['celery_id'] = self.request.id
	context['worker_name'] = os.environ.get('WORKER_NAME', 'unknown')
	opts['context'] = context
	opts['results'] = results
	opts['sync'] = True

	# Initialize task
	sync = not IN_CELERY_WORKER_PROCESS
	task_cls = Task.get_task_class(name)
	task = task_cls(targets, **opts)
	update_state(self, task, force=True)

	# Chunk task if needed
	if task_cls.needs_chunking(targets, sync):
		console.print(Info(message=f'Task {name} requires chunking, breaking into {len(targets)} tasks'))
		return self.replace(break_task(task, opts, targets, results=results))

	# Update state live
	[update_state(self, task) for _ in task]
	update_state(self, task, force=True)

	# Garbage collection to save RAM
	gc.collect()

	return task.results


@app.task
def forward_results(results):
	if isinstance(results, list):
		for ix, item in enumerate(results):
			if isinstance(item, dict) and 'results' in item:
				results[ix] = item['results']
	elif 'results' in results:
		results = results['results']
	results = flatten(results)
	results = deduplicate(results, attr='_uuid')
	console.print(Info(message=f'Forwarding {len(results)} results ...'))
	return results

#--------------#
# Celery utils #
#--------------#


def is_celery_worker_alive():
	"""Check if a Celery worker is available."""
	result = app.control.broadcast('ping', reply=True, limit=1, timeout=1)
	result = bool(result)
	if result:
		console.print(Info(message='Celery worker is available, running remotely'))
	else:
		console.print(Info(message='No Celery worker available, running locally'))
	return result



import shutil

from secator.config import CONFIG
from secator.decorators import task
from secator.runners import Command
from secator.serializers import RegexSerializer
from secator.output_types import Vulnerability, Port, Url, Record, Ip, Tag, Info, Error
from secator.serializers import JSONSerializer


BBOT_MODULES = [
	"affiliates",
	# "ajaxpro",
	"anubisdb",
	"asn",
	"azure_realm",
	"azure_tenant",
	"badsecrets",
	"bevigil",
	"binaryedge",
	# "bucket_aws",
	"bucket_azure",
	"bucket_digitalocean",
	# "bucket_file_enum",
	"bucket_firebase",
	"bucket_google",
	"builtwith",
	"bypass403",
	"c99",
	"censys",
	"certspotter",
	# "chaos",
	"columbus",
	# "credshed",
	# "crobat",
	"crt",
	# "dastardly",
	# "dehashed",
	"digitorus",
	"dnscommonsrv",
	"dnsdumpster",
	# "dnszonetransfer",
	"emailformat",
	"ffuf",
	"ffuf_shortnames",
	# "filedownload",
	"fingerprintx",
	"fullhunt",
	"generic_ssrf",
	"git",
	"telerik",
	# "github_codesearch",
	"github_org",
	"gowitness",
	"hackertarget",
	"host_header",
	"httpx",
	"hunt",
	"hunterio",
	"iis_shortnames",
	# "internetdb",
	# "ip2location",
	"ipneighbor",
	"ipstack",
	"leakix",
	# "masscan",
	# "massdns",
	"myssl",
	# "newsletters",
	# "nmap",
	# "nsec",
	"ntlm",
	"nuclei",
	"oauth",
	"otx",
	"paramminer_cookies",
	"paramminer_getparams",
	"paramminer_headers",
	"passivetotal",
	"pgp",
	# "postman",
	"rapiddns",
	# "riddler",
	"robots",
	"secretsdb",
	"securitytrails",
	"shodan_dns",
	"sitedossier",
	"skymem",
	"smuggler",
	"social",
	"sslcert",
	# "subdomain_hijack",
	"subdomaincenter",
	# "sublist3r",
	"telerik",
	# "threatminer",
	"url_manipulation",
	"urlscan",
	"vhost",
	"viewdns",
	"virustotal",
	# "wafw00f",
	"wappalyzer",
	"wayback",
	"zoomeye"
]
BBOT_PRESETS = [
	'cloud-enum',
	'code-enum',
	'dirbust-heavy',
	'dirbust-light',
	'dotnet-audit',
	'email-enum',
	'iis-shortnames',
	'kitchen-sink',
	'paramminer',
	'spider',
	'subdomain-enum',
	'web-basic',
	'web-screenshots',
	'web-thorough'
]
BBOT_MODULES_STR = ' '.join(BBOT_MODULES)
BBOT_MAP_TYPES = {
	'IP_ADDRESS': Ip,
	'PROTOCOL': Port,
	'OPEN_TCP_PORT': Port,
	'URL': Url,
	'TECHNOLOGY': Tag,
	'ASN': Record,
	'DNS_NAME': Record,
	'WEBSCREENSHOT': Url,
	'VULNERABILITY': Vulnerability,
	'FINDING': Tag
}
BBOT_DESCRIPTION_REGEX = RegexSerializer(
	regex=r'(?P<name>[\w ]+): \[(?P<value>[^\[\]]+)\]',
	findall=True
)


def output_discriminator(self, item):
	_type = item.get('type')
	_message = item.get('message')
	if not _type and _message:
		return Error
	elif _type not in BBOT_MAP_TYPES:
		return None
	return BBOT_MAP_TYPES[_type]


@task()
class bbot(Command):
	"""Multipurpose scanner."""
	cmd = 'bbot -y --allow-deadly --force'
	json_flag = '--json'
	input_flag = '-t'
	file_flag = None
	version_flag = '--help'
	opts = {
		'modules': {'type': str, 'short': 'm', 'default': '', 'help': ','.join(BBOT_MODULES)},
		'presets': {'type': str, 'short': 'ps', 'default': 'kitchen-sink', 'help': ','.join(BBOT_PRESETS), 'shlex': False},
	}
	opt_key_map = {
		'modules': 'm',
		'presets': 'p'
	}
	opt_value_map = {
		'presets': lambda x: ' '.join(x.split(','))
	}
	item_loaders = [JSONSerializer()]
	output_types = [Vulnerability, Port, Url, Record, Ip]
	output_discriminator = output_discriminator
	output_map = {
		Ip: {
			'ip': lambda x: x['data'],
			'host': lambda x: x['data'],
			'alive': lambda x: True,
			'_source': lambda x: 'bbot-' + x['module']
		},
		Tag: {
			'name': 'name',
			'match': lambda x: x['data'].get('url') or x['data'].get('host'),
			'extra_data': 'extra_data',
			'_source': lambda x: 'bbot-' + x['module']
		},
		Url: {
			'url': lambda x: x['data'].get('url') if isinstance(x['data'], dict) else x['data'],
			'host': lambda x: x['resolved_hosts'][0] if 'resolved_hosts' in x else '',
			'status_code': lambda x: bbot.extract_status_code(x),
			'title': lambda x: bbot.extract_title(x),
			'screenshot_path': lambda x: x['data']['path'] if isinstance(x['data'], dict) else '',
			'_source': lambda x: 'bbot-' + x['module']
		},
		Port: {
			'port': lambda x: int(x['data']['port']) if 'port' in x['data'] else int(x['data'].split(':')[-1]),
			'ip': lambda x: [_ for _ in x['resolved_hosts'] if not _.startswith('::')][0],
			'state': lambda x: 'OPEN',
			'service_name': lambda x: x['data']['protocol'] if 'protocol' in x['data'] else '',
			'cpes': lambda x: [],
			'host': lambda x: x['data']['host'] if isinstance(x['data'], dict) else x['data'].split(':')[0],
			'extra_data': 'extra_data',
			'_source': lambda x: 'bbot-' + x['module']
		},
		Vulnerability: {
			'name': 'name',
			'match': lambda x: x['data'].get('url') or x['data']['host'],
			'extra_data': 'extra_data',
			'severity': lambda x: x['data']['severity'].lower()
		},
		Record: {
			'name': 'name',
			'type': 'type',
			'extra_data': 'extra_data'
		},
		Error: {
			'message': 'message'
		}
	}
	install_pre = {
		'apk': ['python3-dev', 'linux-headers', 'musl-dev', 'gcc', 'git', 'openssl', 'unzip', 'tar', 'chromium'],
		'*': ['gcc', 'git', 'openssl', 'unzip', 'tar', 'chromium']
	}
	install_cmd = 'pipx install bbot && pipx upgrade bbot'
	install_post = {
		'*': f'rm -fr {CONFIG.dirs.share}/pipx/venvs/bbot/lib/python3.12/site-packages/ansible_collections/*'
	}

	@staticmethod
	def on_json_loaded(self, item):
		_type = item.get('type')

		if not _type:
			yield item
			return

		# Set scan name and base path for output
		if _type == 'SCAN':
			self.scan_config = item['data']
			return

		if _type not in BBOT_MAP_TYPES:
			self._print(f'[bold orange3]Found unsupported bbot type: {_type}.[/] [bold green]Skipping.[/]', rich=True)
			return

		if isinstance(item['data'], str):
			item['name'] = item['data']
			yield item
			return

		item['extra_data'] = item['data']

		# Parse bbot description into extra_data
		description = item['data'].get('description')
		if description:
			del item['data']['description']
			match = BBOT_DESCRIPTION_REGEX.run(description)
			for chunk in match:
				key, val = tuple([c.strip() for c in chunk])
				if ',' in val:
					val = val.split(',')
				key = '_'.join(key.split(' ')).lower()
				item['extra_data'][key] = val

		# Set technology as name for Tag
		if item['type'] == 'TECHNOLOGY':
			item['name'] = item['data']['technology']
			del item['data']['technology']

		# If 'name' key is present in 'data', set it as name
		elif 'name' in item['data'].keys():
			item['name'] = item['data']['name']
			del item['data']['name']

		# If 'name' key is present in 'extra_data', set it as name
		elif 'extra_data' in item and 'name' in item['extra_data'].keys():
			item['name'] = item['extra_data']['name']
			del item['extra_data']['name']

		# If 'discovery_context' and no name set yet, set it as name
		else:
			item['name'] = item['discovery_context']

		# If a screenshot was saved, move it to secator output folder
		if item['type'] == 'WEBSCREENSHOT':
			from pathlib import Path
			path = Path.home() / '.bbot' / 'scans' / self.scan_config['name'] / item['data']['path']
			name = path.as_posix().split('/')[-1]
			secator_path = f'{self.reports_folder}/.outputs/{name}'
			yield Info(f'Copying screenshot {path} to {secator_path}')
			shutil.copy(path, secator_path)
			item['data']['path'] = secator_path

		yield item

	@staticmethod
	def extract_title(item):
		for tag in item['tags']:
			if 'http-title' in tag:
				title = ' '.join(tag.split('-')[2:])
				return title
		return ''

	@staticmethod
	def extract_status_code(item):
		for tag in item['tags']:
			if 'status-' in tag:
				return int([tag.split('-')[-1]][0])
		return 0



from secator.runners import Command
from secator.decorators import task
from secator.output_types import Vulnerability


@task()
class ls(Command):
    cmd = 'ls -al'
    output_types = [Vulnerability]
    output_map = {
        Vulnerability: {}
	}

    @staticmethod
    def item_loader(self, line):
        fields = ['permissions', 'link_count', 'owner', 'group', 'size', 'month', 'day', 'hour', 'path']
        result = [c for c in line.split(' ') if c]
        if len(result) != len(fields):
            return None
        data = {}
        for ix, value in enumerate(result):
            data[fields[ix]] = value

        # Output vulnerabilities
        permissions = data['permissions']
        path = data['path']
        full_path = f'{self.inputs[0]}/{path}'
        if permissions[-2] == 'w':  # found a vulnerability !
            yield Vulnerability(
                name='World-writeable path',
                severity='high',
                confidence='high',
                provider='ls',
                matched_at=full_path,
                extra_data={k: v for k, v in data.items() if k != 'path'}
            )



import os
import csv
import yaml

from secator.config import CONFIG
from secator.exporters._base import Exporter
from secator.output_types import Info, Error
from secator.rich import console
from secator.utils import pluralize


class GdriveExporter(Exporter):
	def send(self):
		import gspread
		ws = self.report.workspace_name
		info = self.report.data['info']
		title = self.report.data['info']['title']
		sheet_title = f'{self.report.data["info"]["title"]}_{self.report.timestamp}'
		results = self.report.data['results']
		if not CONFIG.addons.gdrive.credentials_path:
			error = Error('Missing CONFIG.addons.gdrive.credentials_path to save to Google Sheets')
			console.print(error)
			return
		if not CONFIG.addons.gdrive.drive_parent_folder_id:
			error = Error('Missing CONFIG.addons.gdrive.drive_parent_folder_id to save to Google Sheets.')
			console.print(error)
			return
		client = gspread.service_account(CONFIG.addons.gdrive.credentials_path)

		# Create workspace folder if it doesn't exist
		folder_id = self.get_folder_by_name(ws, parent_id=CONFIG.addons.gdrive.drive_parent_folder_id)
		if ws and not folder_id:
			folder_id = self.create_folder(
				folder_name=ws,
				parent_id=CONFIG.addons.gdrive.drive_parent_folder_id)

		# Create worksheet
		sheet = client.create(title, folder_id=folder_id)

		# Add options worksheet for input data
		info = self.report.data['info']
		info['targets'] = '\n'.join(info['targets'])
		info['opts'] = yaml.dump(info['opts'])
		keys = [k.replace('_', ' ').upper() for k in list(info.keys())]
		ws = sheet.add_worksheet('OPTIONS', rows=2, cols=len(keys))
		sheet.values_update(
			ws.title,
			params={'valueInputOption': 'USER_ENTERED'},
			body={'values': [keys, list(info.values())]}
		)

		# Add one worksheet per output type
		for output_type, items in results.items():
			items = [i.toDict() for i in items]
			if not items:
				continue
			keys = [
				k.replace('_', ' ').upper()
				for k in list(items[0].keys())
			]
			csv_path = f'{self.report.output_folder}/report_{output_type}.csv'
			if not os.path.exists(csv_path):
				error = Error(
					f'Unable to find CSV at {csv_path}. For Google sheets reports, please enable CSV reports as well.')
				console.print(error)
				return
			sheet_title = pluralize(output_type).upper()
			ws = sheet.add_worksheet(sheet_title, rows=len(items), cols=len(keys))
			with open(csv_path, 'r') as f:
				data = csv.reader(f)
				data = list(data)
				data[0] = [
					k.replace('_', ' ').upper()
					for k in data[0]
				]
				sheet.values_update(
					ws.title,
					params={'valueInputOption': 'USER_ENTERED'},
					body={'values': data}
				)

		# Delete 'default' worksheet
		ws = sheet.get_worksheet(0)
		sheet.del_worksheet(ws)

		info = Info(message=f'Saved Google Sheets reports to [u magenta]{sheet.url}')
		console.print(info)

	def create_folder(self, folder_name, parent_id=None):
		from googleapiclient.discovery import build
		from google.oauth2 import service_account
		creds = service_account.Credentials.from_service_account_file(CONFIG.addons.gdrive.credentials_path)
		service = build('drive', 'v3', credentials=creds)
		body = {
			'name': folder_name,
			'mimeType': "application/vnd.google-apps.folder"
		}
		if parent_id:
			body['parents'] = [parent_id]
		folder = service.files().create(body=body, fields='id').execute()
		return folder['id']

	def list_folders(self, parent_id):
		from googleapiclient.discovery import build
		from google.oauth2 import service_account
		creds = service_account.Credentials.from_service_account_file(CONFIG.addons.gdrive.credentials_path)
		service = build('drive', 'v3', credentials=creds)
		driveid = service.files().get(fileId='root').execute()['id']
		response = service.files().list(
			q=f"'{parent_id}' in parents and mimeType='application/vnd.google-apps.folder'",
			driveId=driveid,
			corpora='drive',
			includeItemsFromAllDrives=True,
			supportsAllDrives=True
		).execute()
		return response

	def get_folder_by_name(self, name, parent_id=None):
		response = self.list_folders(parent_id=parent_id)
		existing = [i for i in response['files'] if i['name'] == name]
		if existing:
			return existing[0]['id']
		return None



from secator.exporters._base import Exporter
from secator.rich import console


class ConsoleExporter(Exporter):
	def send(self):
		results = self.report.data['results']
		for items in results.values():
			for item in items:
				console.print(item)



from secator.config import CONFIG
from secator.runners import Runner
from secator.utils import discover_tasks
from secator.celery_utils import CeleryData
from secator.output_types import Info


class Task(Runner):
	default_exporters = CONFIG.tasks.exporters
	enable_hooks = False

	@classmethod
	def delay(cls, *args, **kwargs):
		from secator.celery import run_task
		return run_task.apply_async(kwargs={'args': args, 'kwargs': kwargs}, queue='celery')

	def yielder(self):
		"""Run task.

		Yields:
			secator.output_types.OutputType: Secator output type.
		"""
		# Get task class
		task_cls = Task.get_task_class(self.config.name)

		# Run opts
		run_opts = self.run_opts.copy()
		run_opts.pop('output', None)
		run_opts.pop('no_poll', False)

		# Set task output types
		self.output_types = task_cls.output_types
		self.enable_duplicate_check = False

		# Get hooks
		hooks = {task_cls: self.hooks}
		run_opts['hooks'] = hooks
		run_opts['context'] = self.context

		# Run task
		if self.sync:
			self.print_item = False
			result = task_cls.si(self.inputs, **run_opts)
			results = result.apply().get()
		else:
			self.celery_result = task_cls.delay(self.inputs, **run_opts)
			self.add_subtask(self.celery_result.id, self.config.name, self.config.description or '')
			yield Info(
				message=f'Celery task created: {self.celery_result.id}',
				task_id=self.celery_result.id
			)
			if self.no_poll:
				return
			results = CeleryData.iter_results(
				self.celery_result,
				ids_map=self.celery_ids_map,
				description=True,
				print_remote_info=False,
				print_remote_title=f'[bold gold3]{self.__class__.__name__.capitalize()}[/] [bold magenta]{self.name}[/] results')

		# Yield task results
		yield from results

	@staticmethod
	def get_task_class(name):
		"""Get task class from a name.

		Args:
			name (str): Task name.
		"""
		if '/' in name:
			name = name.split('/')[0]
		tasks_classes = discover_tasks()
		for task_cls in tasks_classes:
			if task_cls.__name__ == name:
				return task_cls
		raise ValueError(f'Task {name} not found. Aborting.')



from secator.config import CONFIG
from secator.decorators import task
from secator.definitions import (DELAY, FOLLOW_REDIRECT, HEADER,
							   OPT_NOT_SUPPORTED, PROXY, RATE_LIMIT, RETRIES,
							   THREADS, TIMEOUT, USER_AGENT)
from secator.output_types import Vulnerability
from secator.tasks._categories import VulnCode


@task()
class grype(VulnCode):
	"""Vulnerability scanner for container images and filesystems."""
	cmd = 'grype --quiet'
	input_flag = ''
	file_flag = OPT_NOT_SUPPORTED
	json_flag = None
	opt_prefix = '--'
	opt_key_map = {
		HEADER: OPT_NOT_SUPPORTED,
		DELAY: OPT_NOT_SUPPORTED,
		FOLLOW_REDIRECT: OPT_NOT_SUPPORTED,
		PROXY: OPT_NOT_SUPPORTED,
		RATE_LIMIT: OPT_NOT_SUPPORTED,
		RETRIES: OPT_NOT_SUPPORTED,
		THREADS: OPT_NOT_SUPPORTED,
		TIMEOUT: OPT_NOT_SUPPORTED,
		USER_AGENT: OPT_NOT_SUPPORTED
	}
	output_types = [Vulnerability]
	install_pre = {
		'*': ['curl']
	}
	install_cmd = (
		f'curl -sSfL https://raw.githubusercontent.com/anchore/grype/main/install.sh | sh -s -- -b {CONFIG.dirs.bin}'
	)
	install_github_handle = 'anchore/grype'

	@staticmethod
	def item_loader(self, line):
		"""Load vulnerabilty dicts from grype line output."""
		split = [i for i in line.split('  ') if i]
		if len(split) not in [5, 6] or split[0] == 'NAME':
			return
		versions_fixed = None
		if len(split) == 5:  # no version fixed
			product, version, product_type, vuln_id, severity = tuple(split)
		elif len(split) == 6:
			product, version, versions_fixed, product_type, vuln_id, severity = tuple(split)
		extra_data = {
			'lang': product_type,
			'product': product,
			'version': version,
		}
		if versions_fixed:
			extra_data['versions_fixed'] = [c.strip() for c in versions_fixed.split(', ')]
		data = {
			'id': vuln_id,
			'name': vuln_id,
			'matched_at': self.inputs[0],
			'confidence': 'medium',
			'severity': severity.lower(),
			'provider': 'grype',
			'cvss_score': -1,
			'tags': [],
		}
		if vuln_id.startswith('GHSA'):
			data['provider'] = 'github.com'
			data['references'] = [f'https://github.com/advisories/{vuln_id}']
			vuln = VulnCode.lookup_ghsa(vuln_id)
			if vuln:
				data.update(vuln)
				data['severity'] = data['severity'] or severity.lower()
				extra_data['ghsa_id'] = vuln_id
		elif vuln_id.startswith('CVE'):
			vuln = VulnCode.lookup_cve(vuln_id)
			if vuln:
				data.update(vuln)
				data['severity'] = data['severity'] or severity.lower()
		data['extra_data'] = extra_data
		yield data



import time
from dataclasses import dataclass, field

from secator.definitions import HOST, NAME, TYPE
from secator.output_types import OutputType
from secator.utils import rich_to_ansi, rich_escape as _s


@dataclass
class Record(OutputType):
	name: str
	type: str
	host: str = ''
	extra_data: dict = field(default_factory=dict, compare=False)
	_source: str = field(default='', repr=True)
	_type: str = field(default='record', repr=True)
	_timestamp: int = field(default_factory=lambda: time.time(), compare=False)
	_uuid: str = field(default='', repr=True, compare=False)
	_context: dict = field(default_factory=dict, repr=True, compare=False)
	_tagged: bool = field(default=False, repr=True, compare=False)
	_duplicate: bool = field(default=False, repr=True, compare=False)
	_related: list = field(default_factory=list, compare=False)

	_table_fields = [NAME, HOST, TYPE]
	_sort_by = (TYPE, NAME)

	def __str__(self) -> str:
		return self.name

	def __repr__(self) -> str:
		s = rf'🎤 [bold white]{self.name}[/] \[[green]{self.type}[/]]'
		if self.host:
			s += rf' \[[magenta]{self.host}[/]]'
		if self.extra_data:
			s += r' \[[bold yellow]' + ','.join(f'{_s(k)}={_s(v)}' for k, v in self.extra_data.items()) + '[/]]'
		return rich_to_ansi(s)



import csv as _csv

from dataclasses import fields

from secator.exporters._base import Exporter
from secator.rich import console
from secator.output_types import FINDING_TYPES
from secator.output_types import Info


class CsvExporter(Exporter):
	def send(self):
		results = self.report.data['results']
		if not results:
			return
		csv_paths = []

		for output_type, items in results.items():
			output_cls = [o for o in FINDING_TYPES if o._type == output_type][0]
			keys = [o.name for o in fields(output_cls)]
			items = [i.toDict() for i in items]
			if not items:
				continue
			csv_path = f'{self.report.output_folder}/report_{output_type}.csv'
			csv_paths.append(csv_path)
			with open(csv_path, 'w', newline='') as output_file:
				dict_writer = _csv.DictWriter(output_file, keys)
				dict_writer.writeheader()
				dict_writer.writerows(items)

		if len(csv_paths) == 1:
			csv_paths_str = csv_paths[0]
		else:
			csv_paths_str = '\n   • ' + '\n   • '.join(csv_paths)

		info = Info(message=f'Saved CSV reports to {csv_paths_str}')
		console.print(info)



from furl import furl

from secator.decorators import task
from secator.definitions import (CONTENT_LENGTH, DELAY, DEPTH, FILTER_CODES,
							   FILTER_REGEX, FILTER_SIZE, FILTER_WORDS,
							   FOLLOW_REDIRECT, HEADER, MATCH_CODES,
							   MATCH_REGEX, MATCH_SIZE, MATCH_WORDS, METHOD,
							   OPT_NOT_SUPPORTED, PROXY, RATE_LIMIT, RETRIES,
							   STATUS_CODE, THREADS, TIMEOUT, URL, USER_AGENT)
from secator.output_types import Url
from secator.serializers import JSONSerializer
from secator.tasks._categories import HttpCrawler


@task()
class gospider(HttpCrawler):
	"""Fast web spider written in Go."""
	cmd = 'gospider --js'
	file_flag = '-S'
	input_flag = '-s'
	json_flag = '--json'
	opt_prefix = '--'
	opt_key_map = {
		HEADER: 'header',
		DELAY: 'delay',
		DEPTH: 'depth',
		FILTER_CODES: OPT_NOT_SUPPORTED,
		FILTER_REGEX: OPT_NOT_SUPPORTED,
		FILTER_SIZE: OPT_NOT_SUPPORTED,
		FILTER_WORDS: OPT_NOT_SUPPORTED,
		FOLLOW_REDIRECT: 'no-redirect',
		MATCH_CODES: OPT_NOT_SUPPORTED,
		MATCH_REGEX: OPT_NOT_SUPPORTED,
		MATCH_SIZE: OPT_NOT_SUPPORTED,
		MATCH_WORDS: OPT_NOT_SUPPORTED,
		METHOD: OPT_NOT_SUPPORTED,
		PROXY: 'proxy',
		RATE_LIMIT: OPT_NOT_SUPPORTED,
		RETRIES: OPT_NOT_SUPPORTED,
		THREADS: 'threads',
		TIMEOUT: 'timeout',
		USER_AGENT: 'user-agent',
	}
	opt_value_map = {
		FOLLOW_REDIRECT: lambda x: not x,
		DELAY: lambda x: round(x) if isinstance(x, float) else x
	}
	item_loaders = [JSONSerializer()]
	output_map = {
		Url: {
			URL: 'output',
			STATUS_CODE: 'status',
			CONTENT_LENGTH: 'length',
		}
	}
	install_cmd = 'go install -v github.com/jaeles-project/gospider@latest'
	install_github_handle = 'jaeles-project/gospider'
	proxychains = False
	proxy_socks5 = True  # with leaks... https://github.com/jaeles-project/gospider/issues/61
	proxy_http = True  # with leaks... https://github.com/jaeles-project/gospider/issues/61
	profile = 'io'

	@staticmethod
	def validate_item(self, item):
		"""Keep only items that match the same host."""
		if not isinstance(item, dict):
			return False
		try:
			netloc_in = furl(item['input']).netloc
			netloc_out = furl(item['output']).netloc
			if netloc_in != netloc_out:
				return False
		except ValueError:  # gospider returns invalid URLs for output sometimes
			return False
		return True



import os
import json

from secator.decorators import task
from secator.definitions import EMAIL, OUTPUT_PATH
from secator.tasks._categories import OSInt
from secator.output_types import UserAccount, Info, Error


@task()
class h8mail(OSInt):
	"""Email information and password lookup tool."""
	cmd = 'h8mail'
	json_flag = '--json '
	input_flag = '--targets'
	input_type = EMAIL
	file_flag = '-domain'
	version_flag = '--help'
	opt_prefix = '--'
	opts = {
		'config': {'type': str, 'help': 'Configuration file for API keys'},
		'local_breach': {'type': str, 'short': 'lb', 'help': 'Local breach file'}
	}
	install_cmd = 'pipx install h8mail && pipx upgrade h8mail'

	@staticmethod
	def on_start(self):
		output_path = self.get_opt_value(OUTPUT_PATH)
		if not output_path:
			output_path = f'{self.reports_folder}/.outputs/{self.unique_name}.json'
		self.output_path = output_path
		self.cmd = self.cmd.replace('--json', f'--json {self.output_path}')

	@staticmethod
	def on_cmd_done(self):
		if not os.path.exists(self.output_path):
			yield Error(message=f'Could not find JSON results in {self.output_path}')
			return

		yield Info(message=f'JSON results saved to {self.output_path}')
		with open(self.output_path, 'r') as f:
			data = json.load(f)

		targets = data['targets']
		for target in targets:
			email = target['target']
			target_data = target.get('data', [])
			pwn_num = target['pwn_num']
			if not pwn_num > 0:
				continue
			if len(target_data) > 0:
				entries = target_data[0]
				for entry in entries:
					source, site_name = tuple(entry.split(':'))
					yield UserAccount(**{
						"site_name": site_name,
						"username": email.split('@')[0],
						"email": email,
						"extra_data": {
							'source': source
						},
					})
			else:
				yield UserAccount(**{
					"username": email.split('@')[0],
					"email": email,
					"extra_data": {
						'source': self.get_opt_value('local_breach')
					},
				})



import os

from secator.output_types import Error
from secator.utils import deduplicate, debug


def run_extractors(results, opts, inputs=[]):
	"""Run extractors and merge extracted values with option dict.

	Args:
		results (list): List of results.
		opts (dict): Options.
		inputs (list): Original inputs.

	Returns:
		tuple: inputs, options, errors.
	"""
	extractors = {k: v for k, v in opts.items() if k.endswith('_')}
	errors = []
	for key, val in extractors.items():
		key = key.rstrip('_')
		values, err = extract_from_results(results, val)
		errors.extend(err)
		if key == 'targets':
			inputs = deduplicate(values)
		else:
			opts[key] = deduplicate(values)
	return inputs, opts, errors


def extract_from_results(results, extractors):
	"""Extract sub extractors from list of results dict.

	Args:
		results (list): List of dict.
		extractors (list): List of extractors to extract from.

	Returns:
		tuple: List of extracted results (flat), list of errors.
	"""
	extracted_results = []
	errors = []
	if not isinstance(extractors, list):
		extractors = [extractors]
	for extractor in extractors:
		try:
			extracted_results.extend(process_extractor(results, extractor))
		except Exception as e:
			error = Error.from_exception(e)
			errors.append(error)
	return extracted_results, errors


def process_extractor(results, extractor, ctx={}):
	"""Process extractor.

	Args:
		results (list): List of results.
		extractor (dict / str): extractor definition.

	Returns:
		list: List of extracted results.
	"""
	debug('before extract', obj={'results': results, 'extractor': extractor}, sub='extractor')
	if isinstance(extractor, dict):
		_type = extractor['type']
		_field = extractor.get('field')
		_condition = extractor.get('condition', 'True')
	else:
		_type, _field = tuple(extractor.split('.'))
		_condition = 'True'
	items = [
		item for item in results if item._type == _type and eval(_condition)
	]
	if _field:
		_field = '{' + _field + '}' if not _field.startswith('{') else _field
		items = [_field.format(**item.toDict()) for item in items]
	debug('after extract', obj={'items': items}, sub='extractor')
	return items


def get_task_folder_id(path):
	names = []
	if not os.path.exists(path):
		return 0
	for f in os.scandir(path):
		if f.is_dir():
			try:
				int(f.name)
				names.append(int(f.name))
			except ValueError:
				continue
	names.sort()
	if names:
		return names[-1] + 1
	return 0



import os

import yaml

from secator.decorators import task
from secator.definitions import (CONTENT_LENGTH, CONTENT_TYPE, DELAY, DEPTH,
							   FILTER_CODES, FILTER_REGEX, FILTER_SIZE,
							   FILTER_WORDS, FOLLOW_REDIRECT, HEADER,
							   MATCH_CODES, MATCH_REGEX, MATCH_SIZE,
							   MATCH_WORDS, METHOD, OPT_NOT_SUPPORTED, OUTPUT_PATH, PROXY,
							   RATE_LIMIT, RETRIES, STATUS_CODE,
							   THREADS, TIMEOUT, USER_AGENT, WORDLIST)
from secator.output_types import Url, Info, Error
from secator.tasks._categories import HttpFuzzer


@task()
class dirsearch(HttpFuzzer):
	"""Advanced web path brute-forcer."""
	cmd = 'dirsearch'
	input_flag = '-u'
	file_flag = '-l'
	json_flag = '-O json'
	opt_prefix = '--'
	encoding = 'ansi'
	opt_key_map = {
		HEADER: 'header',
		DELAY: 'delay',
		DEPTH: 'max-recursion-depth',
		FILTER_CODES: 'exclude-status',
		FILTER_REGEX: 'exclude-regex',
		FILTER_SIZE: 'exclude-sizes',
		FILTER_WORDS: OPT_NOT_SUPPORTED,
		FOLLOW_REDIRECT: 'follow-redirects',
		MATCH_CODES: 'include-status',
		MATCH_REGEX: OPT_NOT_SUPPORTED,
		MATCH_SIZE: OPT_NOT_SUPPORTED,
		MATCH_WORDS: OPT_NOT_SUPPORTED,
		METHOD: 'http-method',
		PROXY: 'proxy',
		RATE_LIMIT: 'max-rate',
		RETRIES: 'retries',
		THREADS: 'threads',
		TIMEOUT: 'timeout',
		USER_AGENT: 'user-agent',
		WORDLIST: 'wordlists',
	}
	output_map = {
		Url: {
			CONTENT_LENGTH: 'content-length',
			CONTENT_TYPE: 'content-type',
			STATUS_CODE: 'status'
		}
	}
	install_cmd = 'pipx install --force git+https://github.com/maurosoria/dirsearch'
	proxychains = True
	proxy_socks5 = True
	proxy_http = True
	profile = 'io'

	@staticmethod
	def on_init(self):
		self.output_path = self.get_opt_value(OUTPUT_PATH)
		if not self.output_path:
			self.output_path = f'{self.reports_folder}/.outputs/{self.unique_name}.json'
		self.cmd += f' -o {self.output_path}'

	@staticmethod
	def on_cmd_done(self):
		if not os.path.exists(self.output_path):
			yield Error(message=f'Could not find JSON results in {self.output_path}')
			return

		yield Info(message=f'JSON results saved to {self.output_path}')
		with open(self.output_path, 'r') as f:
			results = yaml.safe_load(f.read()).get('results', [])
		yield from results



import logging

from secator.template import TemplateLoader
from secator.config import CONFIG
from secator.runners._base import Runner
from secator.runners._helpers import run_extractors
from secator.runners.workflow import Workflow
from secator.utils import merge_opts

logger = logging.getLogger(__name__)


class Scan(Runner):

	default_exporters = CONFIG.scans.exporters

	@classmethod
	def delay(cls, *args, **kwargs):
		from secator.celery import run_scan
		return run_scan.delay(args=args, kwargs=kwargs)

	def yielder(self):
		"""Run scan.

		Yields:
			secator.output_types.OutputType: Secator output type.
		"""
		scan_opts = self.config.options
		self.print_item = False
		for name, workflow_opts in self.config.workflows.items():

			# Extract opts and and expand target from previous workflows results
			targets, workflow_opts = run_extractors(self.results, workflow_opts or {}, self.inputs)

			# Run workflow
			run_opts = self.run_opts.copy()
			opts = merge_opts(scan_opts, workflow_opts, run_opts)
			workflow = Workflow(
				TemplateLoader(name=f'workflows/{name}'),
				targets,
				results=[],
				run_opts=opts,
				hooks=self._hooks,
				context=self.context.copy())

			# Get results
			yield from workflow



from secator.decorators import task
from secator.definitions import (AUTO_CALIBRATION, CONTENT_LENGTH,
								 CONTENT_TYPE, DELAY, DEPTH, EXTRA_DATA,
								 FILTER_CODES, FILTER_REGEX, FILTER_SIZE,
								 FILTER_WORDS, FOLLOW_REDIRECT, HEADER,
								 MATCH_CODES, MATCH_REGEX, MATCH_SIZE,
								 MATCH_WORDS, METHOD, OPT_NOT_SUPPORTED,
								 PERCENT, PROXY, RATE_LIMIT, RETRIES,
								 STATUS_CODE, THREADS, TIME, TIMEOUT,
								 USER_AGENT, WORDLIST)
from secator.output_types import Progress, Url
from secator.serializers import JSONSerializer, RegexSerializer
from secator.tasks._categories import HttpFuzzer

FFUF_PROGRESS_REGEX = r':: Progress: \[(?P<count>\d+)/(?P<total>\d+)\] :: Job \[\d/\d\] :: (?P<rps>\d+) req/sec :: Duration: \[(?P<duration>[\d:]+)\] :: Errors: (?P<errors>\d+) ::'  # noqa: E501


@task()
class ffuf(HttpFuzzer):
	"""Fast web fuzzer written in Go."""
	cmd = 'ffuf -noninteractive -recursion'
	input_flag = '-u'
	input_chunk_size = 1
	file_flag = None
	json_flag = '-json'
	version_flag = '-V'
	item_loaders = [
		JSONSerializer(strict=True),
		RegexSerializer(FFUF_PROGRESS_REGEX, fields=['count', 'total', 'rps', 'duration', 'errors'])
	]
	opts = {
		AUTO_CALIBRATION: {'is_flag': True, 'short': 'ac', 'help': 'Auto-calibration'},
	}
	opt_key_map = {
		HEADER: 'H',
		DELAY: 'p',
		DEPTH: 'recursion-depth',
		FILTER_CODES: 'fc',
		FILTER_REGEX: 'fr',
		FILTER_SIZE: 'fs',
		FILTER_WORDS: 'fw',
		FOLLOW_REDIRECT: 'r',
		MATCH_CODES: 'mc',
		MATCH_REGEX: 'mr',
		MATCH_SIZE: 'ms',
		MATCH_WORDS: 'mw',
		METHOD: 'X',
		PROXY: 'x',
		RATE_LIMIT: 'rate',
		RETRIES: OPT_NOT_SUPPORTED,
		THREADS: 't',
		TIMEOUT: 'timeout',
		USER_AGENT: OPT_NOT_SUPPORTED,

		# ffuf opts
		WORDLIST: 'w',
		AUTO_CALIBRATION: 'ac',
	}
	output_types = [Url, Progress]
	output_map = {
		Url: {
			STATUS_CODE: 'status',
			CONTENT_LENGTH: 'length',
			CONTENT_TYPE: 'content-type',
			TIME: lambda x: x['duration'] * 10**-9
		},
		Progress: {
			PERCENT: lambda x: int(int(x['count']) * 100 / int(x['total'])),
			EXTRA_DATA: lambda x: {k: v for k, v in x.items() if k not in ['count', 'total', 'errors']}
		},
	}
	encoding = 'ansi'
	install_cmd = 'go install -v github.com/ffuf/ffuf@latest'
	install_github_handle = 'ffuf/ffuf'
	proxychains = False
	proxy_socks5 = True
	proxy_http = True
	profile = 'io'

	@staticmethod
	def on_item(self, item):
		if isinstance(item, Url):
			item.method = self.get_opt_value(METHOD) or 'GET'
		return item



from secator.decorators import task
from secator.definitions import (OPT_PIPE_INPUT, RATE_LIMIT, RETRIES, THREADS)
from secator.output_types import Record, Ip, Subdomain
from secator.output_types.ip import IpProtocol
from secator.tasks._categories import ReconDns
from secator.serializers import JSONSerializer
from secator.utils import extract_domain_info


@task()
class dnsx(ReconDns):
	"""dnsx is a fast and multi-purpose DNS toolkit designed for running various retryabledns library."""
	cmd = 'dnsx -resp -recon'
	json_flag = '-json'
	input_flag = OPT_PIPE_INPUT
	file_flag = OPT_PIPE_INPUT
	output_types = [Record, Ip, Subdomain]
	opt_key_map = {
		RATE_LIMIT: 'rate-limit',
		RETRIES: 'retry',
		THREADS: 'threads',
	}
	opts = {
		'trace': {'is_flag': True, 'default': False, 'help': 'Perform dns tracing'},
		'resolver': {'type': str, 'short': 'r', 'help': 'List of resolvers to use (file or comma separated)'},
		'wildcard_domain': {'type': str, 'short': 'wd', 'help': 'Domain name for wildcard filtering'},
	}
	item_loaders = [JSONSerializer()]
	install_cmd = 'go install -v github.com/projectdiscovery/dnsx/cmd/dnsx@latest'
	install_github_handle = 'projectdiscovery/dnsx'
	profile = 'io'

	@staticmethod
	def on_json_loaded(self, item):
		# Show full DNS response
		quiet = self.get_opt_value('quiet')
		if not quiet:
			all = item['all']
			for line in all:
				yield line
			yield '\n'

		# Loop through record types and yield records
		record_types = ['a', 'aaaa', 'cname', 'mx', 'ns', 'txt', 'srv', 'ptr', 'soa', 'axfr', 'caa']
		host = item['host']
		for _type in record_types:
			values = item.get(_type, [])
			for value in values:
				name = value
				extra_data = {}
				if isinstance(value, dict):
					name = value['name']
					extra_data = {k: v for k, v in value.items() if k != 'name'}
				if _type == 'a':
					yield Ip(
						host=host,
						ip=name,
						protocol=IpProtocol.IPv4
					)
				elif _type == 'aaaa':
					yield Ip(
						host=host,
						ip=name,
						protocol=IpProtocol.IPv6
					)
				elif _type == 'ptr':
					yield Subdomain(
						host=name,
						domain=extract_domain_info(name, domain_only=True)
					)
				yield Record(
					host=host,
					name=name,
					type=_type.upper(),
					extra_data=extra_data
				)



from secator.cli import ALL_SCANS


def generate_class(config):
	from secator.runners import Workflow

	class workflow(Workflow):
		def __init__(self, inputs=[], **run_opts):
			hooks = run_opts.pop('hooks', {})
			results = run_opts.pop('results', [])
			context = run_opts.pop('context', {})
			super().__init__(
				config=config,
				inputs=inputs,
				results=results,
				run_opts=run_opts,
				hooks=hooks,
				context=context)
	return workflow, config.name


DYNAMIC_SCANS = {}
for workflow in ALL_SCANS:
	cls, name = generate_class(workflow)
	DYNAMIC_SCANS[name] = cls

globals().update(DYNAMIC_SCANS)
__all__ = list(DYNAMIC_SCANS)



import json
import re

from secator.decorators import task
from secator.output_types import Url, Progress
from secator.definitions import (
	HEADER, DELAY, FOLLOW_REDIRECT, METHOD, PROXY, RATE_LIMIT, RETRIES, THREADS, TIMEOUT, USER_AGENT,
	DEPTH, MATCH_REGEX, MATCH_SIZE, MATCH_WORDS, FILTER_REGEX, FILTER_CODES, FILTER_SIZE, FILTER_WORDS,
	MATCH_CODES, OPT_NOT_SUPPORTED, URL
)
from secator.serializers import JSONSerializer
from secator.tasks._categories import Http


@task()
class bup(Http):
	"""40X bypasser."""
	cmd = 'bup'
	input_flag = '-u'
	input_type = URL
	json_flag = '--jsonl'
	opt_prefix = '--'
	opts = {
		'spoofport': {'type': int, 'short': 'sp', 'help': 'Port(s) to inject in port-specific headers'},
		'spoofip': {'type': str, 'short': 'si', 'help': 'IP(s) to inject in ip-specific headers'},
		'mode': {'type': str, 'help': 'Bypass modes.'},
	}
	opt_key_map = {
		HEADER: 'header',
		DELAY: OPT_NOT_SUPPORTED,
		FOLLOW_REDIRECT: OPT_NOT_SUPPORTED,
		METHOD: OPT_NOT_SUPPORTED,
		RATE_LIMIT: OPT_NOT_SUPPORTED,
		RETRIES: 'retry',
		THREADS: 'threads',
		TIMEOUT: 'timeout',
		USER_AGENT: OPT_NOT_SUPPORTED,
		DEPTH: OPT_NOT_SUPPORTED,
		MATCH_REGEX: OPT_NOT_SUPPORTED,
		MATCH_SIZE: OPT_NOT_SUPPORTED,
		MATCH_WORDS: OPT_NOT_SUPPORTED,
		FILTER_REGEX: OPT_NOT_SUPPORTED,
		FILTER_CODES: OPT_NOT_SUPPORTED,
		FILTER_SIZE: OPT_NOT_SUPPORTED,
		FILTER_WORDS: OPT_NOT_SUPPORTED,
		MATCH_CODES: OPT_NOT_SUPPORTED,
		PROXY: 'proxy',
	}
	item_loaders = [JSONSerializer()]
	output_types = [Url, Progress]
	output_map = {
		Url: {
			'url': 'request_url',
			'method': lambda x: bup.method_extractor(x),
			'headers': lambda x: bup.headers_extractor(x),
			'status_code': 'response_status_code',
			'content_type': 'response_content_type',
			'content_length': 'response_content_length',
			'title': 'response_title',
			'server': 'response_server_type',
			'lines': 'response_lines_count',
			'words': 'response_words_count',
			'stored_response_path': 'response_html_filename',
		}
	}
	install_cmd = 'pipx install bypass-url-parser && pipx upgrade bypass-url-parser'

	@staticmethod
	def on_init(self):
		self.cmd += f' -o {self.reports_folder}/.outputs/response'

	@staticmethod
	def on_line(self, line):
		if 'Doing' in line:
			progress_indicator = line.split(':')[-1]
			current, total = tuple([int(c.strip()) for c in progress_indicator.split('/')])
			return json.dumps({"duration": "unknown", "percent": int((current / total) * 100)})
		elif 'batcat' in line:  # ignore batcat lines as they're loaded as JSON
			return None
		return line

	@staticmethod
	def method_extractor(item):
		payload = item['request_curl_payload']
		match = re.match(r'-X\s+(\w+)', payload)
		if match:
			return match.group(1)
		return 'GET'

	@staticmethod
	def headers_extractor(item):
		headers_list = item['response_headers'].split('\n')[1:]
		headers = {}
		for header in headers_list:
			split_headers = header.split(':')
			key = split_headers[0]
			value = ':'.join(split_headers[1:])
			headers[key] = value
		return headers



from secator.exporters._base import Exporter
from secator.output_types import Info
from secator.rich import console


class TxtExporter(Exporter):
	def send(self):
		results = self.report.data['results']
		if not results:
			return
		txt_paths = []

		for output_type, items in results.items():
			items = [str(i) for i in items]
			if not items:
				continue
			txt_path = f'{self.report.output_folder}/report_{output_type}.txt'
			with open(txt_path, 'w') as f:
				f.write('\n'.join(items))
			txt_paths.append(txt_path)

		if len(txt_paths) == 1:
			txt_paths_str = txt_paths[0]
		else:
			txt_paths_str = '\n   • ' + '\n   • '.join(txt_paths)

		info = Info(f'Saved TXT reports to {txt_paths_str}')
		console.print(info)



from secator.decorators import task
from secator.definitions import (DELAY, DOMAIN, OPT_NOT_SUPPORTED, PROXY,
							   RATE_LIMIT, RETRIES, THREADS, TIMEOUT)
from secator.output_types import Subdomain
from secator.serializers import JSONSerializer
from secator.tasks._categories import ReconDns


@task()
class subfinder(ReconDns):
	"""Fast passive subdomain enumeration tool."""
	cmd = 'subfinder -silent -cs'
	file_flag = '-dL'
	input_flag = '-d'
	json_flag = '-json'
	opt_key_map = {
		DELAY: OPT_NOT_SUPPORTED,
		PROXY: 'proxy',
		RATE_LIMIT: 'rate-limit',
		RETRIES: OPT_NOT_SUPPORTED,
		TIMEOUT: 'timeout',
		THREADS: 't'
	}
	opt_value_map = {
		PROXY: lambda x: x.replace('http://', '').replace('https://', '') if x else None
	}
	item_loaders = [JSONSerializer()]
	output_map = {
		Subdomain: {
			DOMAIN: 'input',
		}
	}
	output_types = [Subdomain]
	install_cmd = 'go install -v github.com/projectdiscovery/subfinder/v2/cmd/subfinder@latest'
	install_github_handle = 'projectdiscovery/subfinder'
	proxychains = False
	proxy_http = True
	proxy_socks5 = False
	profile = 'io'

	@staticmethod
	def validate_item(self, item):
		if isinstance(item, dict):
			return item['input'] != 'localhost'
		return True



import time
from dataclasses import dataclass, field

from secator.definitions import SITE_NAME, URL, USERNAME
from secator.output_types import OutputType
from secator.utils import rich_to_ansi, rich_escape as _s


@dataclass
class UserAccount(OutputType):
	username: str
	url: str = ''
	email: str = ''
	site_name: str = ''
	extra_data: dict = field(default_factory=dict, compare=False)
	_source: str = field(default='', repr=True)
	_type: str = field(default='user_account', repr=True)
	_timestamp: int = field(default_factory=lambda: time.time(), compare=False)
	_uuid: str = field(default='', repr=True, compare=False)
	_context: dict = field(default_factory=dict, repr=True, compare=False)
	_tagged: bool = field(default=False, repr=True, compare=False)
	_duplicate: bool = field(default=False, repr=True, compare=False)
	_related: list = field(default_factory=list, compare=False)

	_table_fields = [SITE_NAME, USERNAME, URL]
	_sort_by = (URL, USERNAME)

	def __str__(self) -> str:
		return self.url

	def __repr__(self) -> str:
		s = f'👤 [green]{_s(self.username)}[/]'
		if self.email:
			s += rf' \[[bold yellow]{_s(self.email)}[/]]'
		if self.site_name:
			s += rf' \[[bold blue]{self.site_name}[/]]'
		if self.url:
			s += rf' \[[white]{_s(self.url)}[/]]'
		if self.extra_data:
			s += r' \[[bold yellow]' + _s(', '.join(f'{k}:{v}' for k, v in self.extra_data.items()) + '[/]]')
		return rich_to_ansi(s)



"""Attack tasks."""

import logging

from rich.panel import Panel

from secator.config import CONFIG
from secator.decorators import task
from secator.definitions import (DELAY, FOLLOW_REDIRECT, HEADER, HOST, OPT_NOT_SUPPORTED, PROXY, RATE_LIMIT, RETRIES,
								 THREADS, TIMEOUT, USER_AGENT)
from secator.tasks._categories import VulnMulti
from secator.utils import get_file_timestamp

logger = logging.getLogger(__name__)


@task()
class msfconsole(VulnMulti):
	"""CLI to access and work with the Metasploit Framework."""
	cmd = 'msfconsole --quiet'
	version_flag = OPT_NOT_SUPPORTED
	input_type = HOST
	input_chunk_size = 1
	output_types = []
	opt_prefix = '--'
	opts = {
		'resource': {'type': str, 'help': 'Metasploit resource script.', 'short': 'r'},
		'execute_command': {'type': str, 'help': 'Metasploit command.', 'short': 'x'},
		'environment': {'type': str, 'help': 'Environment variables string KEY=VALUE.', 'short': 'e'}
	}
	opt_key_map = {
		'x': 'execute_command',
		'r': 'resource',
		HEADER: OPT_NOT_SUPPORTED,
		DELAY: OPT_NOT_SUPPORTED,
		FOLLOW_REDIRECT: OPT_NOT_SUPPORTED,
		PROXY: OPT_NOT_SUPPORTED,
		RATE_LIMIT: OPT_NOT_SUPPORTED,
		RETRIES: OPT_NOT_SUPPORTED,
		THREADS: OPT_NOT_SUPPORTED,
		TIMEOUT: OPT_NOT_SUPPORTED,
		USER_AGENT: OPT_NOT_SUPPORTED,
	}
	encoding = 'ansi'
	ignore_return_code = True
	install_pre = {
		'apt|apk': ['libpq-dev', 'libpcap-dev', 'libffi-dev', 'g++', 'make'],
		'pacman': ['ruby-erb', 'postgresql-libs', 'make'],
		'yum|zypper': ['postgresql-devel', 'make'],
	}
	install_cmd = (
		f'git clone --depth 1 --single-branch https://github.com/rapid7/metasploit-framework.git {CONFIG.dirs.share}/metasploit-framework || true && '  # noqa: E501
		f'cd {CONFIG.dirs.share}/metasploit-framework && '
		f'gem install bundler --user-install -n {CONFIG.dirs.bin} && '
		f'bundle config set --local path "{CONFIG.dirs.share}" && '
		'bundle lock --normalize-platforms &&'
		'bundle install && '
		f'ln -sf $HOME/.local/share/metasploit-framework/msfconsole {CONFIG.dirs.bin}/msfconsole'
	)

	@staticmethod
	def on_init(self):
		command = self.get_opt_value('execute_command')
		script_path = self.get_opt_value('resource')
		environment = self.run_opts.pop('environment', '')
		env_vars = {}
		if environment:
			env_vars = dict(map(lambda x: x.split('='), environment.strip().split(',')))
		env_vars['RHOST'] = self.inputs[0]
		env_vars['RHOSTS'] = self.inputs[0]

		# Passing msfconsole command directly, simply add RHOST / RHOSTS from host input and run then exit
		if command:
			self.run_opts['msfconsole.execute_command'] = (
				f'setg RHOST {self.inputs[0]}; '
				f'setg RHOSTS {self.inputs[0]}; '
				f'{command.format(**env_vars)}; '
				f'exit;'
			)

		# Passing resource script, replace vars inside by our environment variables if any, write to a temp file, and
		# pass this temp file instead of the original one.
		elif script_path:

			# Read from original resource script
			with open(script_path, 'r') as f:
				content = f.read().replace('exit', '') + 'exit'

			# Make a copy and replace vars inside by env vars passed on the CLI
			timestr = get_file_timestamp()
			out_path = f'{self.reports_folder}/.inputs/msfconsole_{timestr}.rc'
			logger.debug(
				f'Writing formatted resource script to new temp file {out_path}'
			)
			with open(out_path, 'w') as f:
				content = content.format(**env_vars)
				f.write(content)

			script_name = script_path.split('/')[-1]
			self._print(Panel(content, title=f'[bold magenta]{script_name}', expand=False), rich=True)

			# Override original command with new resource script
			self.run_opts['msfconsole.resource'] = out_path

		# Nothing passed, error out
		else:
			raise ValueError('At least one of "execute_command" or "resource" must be passed.')


# TODO: This is better as it goes through an RPC API to communicate with
# metasploit rpc server, but it does not give any output.
# Seems like output is available only in Metasploit Pro, so keeping this in case
# we add support for it later.
#
# from pymetasploit3.msfrpc import MsfRpcClient
# class msfrpcd():
#
#     opts = {
#         'uri': {'type': str, 'default': '/api/', 'help': 'msfrpcd API uri'},
#         'port': {'type': int, 'default': 55553, 'help': 'msfrpcd port'},
#         'server': {'type': str, 'default': 'localhost', 'help': 'msfrpcd host'},
#         'token': {'type': str, 'help': 'msfrpcd token'},
#         'username': {'type': str, 'default': 'msf', 'help': 'msfrpcd username'},
#         'password': {'type': str, 'default': 'test', 'help': 'msfrpcd password'},
#         'module': {'type': str, 'required': True, 'help': 'Metasploit module to run'}
#     }
#
#     def __init__(self, input, ctx={}, **run_opts):
#         self.module = run_opts.pop('module')
#         pw = run_opts.pop('password')
#         self.run_opts = run_opts
#         self.RHOST = input
#         self.RHOSTS = input
#         self.LHOST = self.get_lhost()
#         # self.start_msgrpc()
#         self.client = MsfRpcClient(pw, ssl=True, **run_opts)
#
#     # def start_msgrpc(self):
#     #     code, out = Command.execute(f'msfrpcd -P {self.password}')
#     #     logger.info(out)
#
#     def get_lhost(self):
#         try:
#             u = miniupnpc.UPnP()
#             u.discoverdelay = 200
#             u.discover()
#             u.selectigd()
#             return u.externalipaddress()
#         except Exception:
#             return 'localhost'
#
#     def run(self):
#         """Run a metasploit module.
#
#         Args:
#             modtype: Module type amongst 'auxiliary', 'exploit', 'post',
#                 'encoder', 'nop', 'payload'.
#             modname: Module name e.g 'auxiliary/scanner/ftp/ftp_version
#             kwargs (dict): Module kwargs e.g RHOSTS, LHOST
#         Returns:
#             dict: Job results.
#         """
#         modtype = self.module.split('/')[0].rstrip('s')
#         job = self.client.modules.execute(
# 			modtype,
# 			self.module,
# 			RHOST=self.RHOST,
# 			RHOSTS=self.RHOSTS,
# 			LHOST=self.LHOST)
#         if job.get('error', False):
#             logger.error(job['error_message'])
#         job_info = self.client.jobs.info_by_uuid(job['uuid'])
#         while (job_info['status'] in ['running', 'ready']):
#             job_info = self.client.jobs.info_by_uuid(job['uuid'])
#         job_info.update(job)
#         print(type(job_info['result']['127.0.0.1']))
#         return job_info



import time
from dataclasses import dataclass, field

from secator.output_types import OutputType
from secator.utils import rich_to_ansi, rich_escape as _s


@dataclass
class Target(OutputType):
	name: str
	_source: str = field(default='', repr=True)
	_type: str = field(default='target', repr=True)
	_timestamp: int = field(default_factory=lambda: time.time(), compare=False)
	_uuid: str = field(default='', repr=True, compare=False)
	_context: dict = field(default_factory=dict, repr=True, compare=False)
	_tagged: bool = field(default=False, repr=True, compare=False)
	_duplicate: bool = field(default=False, repr=True, compare=False)
	_related: list = field(default_factory=list, compare=False)

	_table_fields = [
		'name',
	]
	_sort_by = ('name',)

	def __str__(self):
		return self.name

	def __repr__(self):
		s = f'🎯 {_s(self.name)}'
		return rich_to_ansi(s)
